/**
 * Pikku esimerkki deadlockista.
 *
 * Olennaisesti sama esimerkki kuin vuoden 2015 vappusimulaattori,
 * mutta sittemmin on käsitelty vasta vapun jälkeen, joten simuloidaan
 * päivämäärästä riippumattomasti kahvihuonetta.
 *
 *  Tehdään kaksi säiettä, jotka juovat kahvia ja syövät keksejä.
 *  Molemmat resurssit ovat yhteisessä pöydässä, joten kahvipannu ja
 *  keksipurkki täytyy lukita, että yhdistetty kahvinautinto onnistuu.
 *
 * Tässä esimerkissä kahvihetki ei jumiudu, koska kahvin lukitseminen
 * tehdään molemmissa säikeissä ensimmäisenä.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <pthread.h>

#define KAHVIANNOSTEN_TARVE_LKM 1000

uint64_t kahvia_ltr = 1000;
uint64_t keksit_kpl = 1000;

pthread_mutex_t mutex_kahvi = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mutex_keksit = PTHREAD_MUTEX_INITIALIZER;

bool juo_kahvia(const char *juoja){
    if (kahvia_ltr == 0){
        fprintf(stderr, "Ou nou! Kahvi on loppu!\n");
        return false;
    }
    fprintf(stdout, "%s: Yksi litra kahvia juotu.\n", juoja);
    kahvia_ltr--;
    return true;
}

bool ota_keksi(const char *ottaja){
    if (keksit_kpl == 0){
        fprintf(stderr, "Ou nou! Keksit on loppu!\n");
        return false;
    }
    fprintf(stdout, "%s: Yksi keksi syöty.\n", ottaja);
    keksit_kpl--;
    return true;
}

void * paavo(void *v) {
    const char idstr[] = "Paavo";
    int a = 0;
    for (int i = 1; i <= KAHVIANNOSTEN_TARVE_LKM; i++){
        pthread_mutex_lock(&mutex_kahvi);
        pthread_mutex_lock(&mutex_keksit);
        if (juo_kahvia(idstr) && ota_keksi(idstr)){
            printf("%s: Nams, olipa hyvää. Olen saanut %d annosta.\n",idstr,++a);
        } else {
            printf("%s: Hitsi, mulle ei riittänyt! Tänään ei herätä...\n", idstr);
        }
        pthread_mutex_unlock(&mutex_kahvi);
        pthread_mutex_unlock(&mutex_keksit);
    }
    printf("%s: %s Sain yhteensä %d täyttä annosta.\n",
           idstr,(a>0)?"Nams, olipa hyvää.":"",a);
    return NULL;
}

void * ari(void *v) {
    const char idstr[] = "Ari";
    int a = 0;
    for (int i = 1; i <= KAHVIANNOSTEN_TARVE_LKM; i++){
        pthread_mutex_lock(&mutex_kahvi);
        pthread_mutex_lock(&mutex_keksit);
        if (juo_kahvia(idstr) && ota_keksi(idstr)){
            printf("%s: Nams, olipa hyvää. Olen saanut %d annosta.\n",idstr,++a);
        } else {
            printf("%s: Hitsi, mulle ei riittänyt! Tänään ei herätä...\n", idstr);
        }
        pthread_mutex_unlock(&mutex_kahvi);
        pthread_mutex_unlock(&mutex_keksit);
    }
    printf("%s: %s Sain yhteensä %d täyttä annosta.\n",
           idstr,(a>0)?"Nams, olipa hyvää.":"",a);
    return NULL;
}


int main(int argc, char *argv[]) {
    int n;
    pthread_t saieA, saieB;

    if (pthread_create(&saieA, NULL, paavo, NULL) != 0) {
        fprintf(stderr, "Säikeen luonti ei onnistunut.\n");
        exit(1);
    }
    if (pthread_create(&saieB, NULL, ari, NULL) != 0) {
        fprintf(stderr, "Säikeen luonti ei onnistunut.\n");
        exit(1);
    }

    if (pthread_join(saieA, NULL) != 0) {
        fprintf(stderr, "Säikeen A odottelussa tuli ongelmia.\n");
    };
    if (pthread_join(saieB, NULL) != 0) {
        fprintf(stderr, "Säikeen B odottelussa tuli ongelmia.\n");
    };

    printf("Kahvia jäljellä %" PRId64 "; keksejä jäljellä %" PRId64 "\n",
           kahvia_ltr, keksit_kpl);

    return 0;
}
